from typing import Dict, Optional

from pyspark.sql import DataFrame
from pyspark.sql.functions import countDistinct
from pyspark.sql.types import (
    ArrayType,
    BooleanType,
    DateType,
    NumericType,
    StringType,
    TimestampType,
)

from aporia.core.context import get_context
from aporia.core.errors import AporiaError, handle_error
from aporia.core.types.field import FieldType

DATA_SAMPLE_SIZE = 10000
UNIQUE_RATIO = 0.25
UNIQUE_VALUES_THRESHOLD = 50


def infer_schema_from_pyspark_dataframe(data: DataFrame) -> Optional[Dict[str, str]]:
    """Infers model version schema from a PySpark DataFrame.

    Field names and types are inferred from column names and types.

    Args:
        data: PySpark DataFrame

    Returns:
        A schema describing the data, as required by the create_model_version function.

    Notes:
        * The field types are inferred using the following logic, based on the data schema:
            * Boolean data type -> boolean field
            * Datetime data type -> datetime field
            * Array data type with numeric elements -> vector field
            * Numeric data type:
                * At most 50 unique values, at most 25% of the values are unique -> categorical field
                * Otherwise -> numeric field
            * String data type:
                * At least 50 unique values, at least 25% of the values are unique -> text field
                * Otherwise -> string field
        * If data contains a column with a type that doesn't match any of the rules
          described above, an error will be raised.
        * If data is a large dataset (> 10000 rows), a sample of the data will be used to infer the schema.

    See Also:
        * <https://spark.apache.org/docs/latest/sql-ref-datatypes.html#data-types>
    """
    context = get_context()
    schema = {}
    # To avoid long computations, use a sample of the data
    data_sample = data.sample(fraction=min(DATA_SAMPLE_SIZE / data.count(), 1.0))
    total_count = data_sample.count()
    try:
        for column in data.schema:
            unique_count = 0
            # Numeric and String columns might be categorical/text fields, fetch the
            # number of unique values to apply our heuristic later
            if isinstance(column.dataType, (NumericType, StringType)):
                unique_count = (
                    data_sample.select(column.name).agg(countDistinct(column.name)).collect()[0][0]
                )

            if isinstance(column.dataType, BooleanType):
                field_type = FieldType.BOOLEAN
            elif isinstance(column.dataType, (TimestampType, DateType)):
                field_type = FieldType.DATETIME
            elif isinstance(column.dataType, ArrayType) and isinstance(
                column.dataType.elementType, NumericType
            ):
                field_type = FieldType.VECTOR
            elif isinstance(column.dataType, NumericType):
                # See the heuristic described in the function notes
                if (
                    total_count > 0
                    and unique_count < UNIQUE_VALUES_THRESHOLD
                    and unique_count < total_count * UNIQUE_RATIO
                ):
                    field_type = FieldType.CATEGORICAL
                else:
                    field_type = FieldType.NUMERIC
            elif isinstance(column.dataType, StringType):
                # See the heuristic described in the function notes
                if (
                    total_count > 0
                    and unique_count > UNIQUE_VALUES_THRESHOLD
                    and unique_count > total_count * UNIQUE_RATIO
                ):
                    field_type = FieldType.TEXT
                else:
                    field_type = FieldType.STRING
            else:
                raise AporiaError(
                    "the type {} of column {} is not supported".format(column.dataType, column.name)
                )

            schema[column.name] = field_type.value

        return schema

    except Exception as err:
        handle_error(
            message_format="Inferring schema from PySpark dataframe failed, {}",
            verbose=context.config.verbose,
            throw_errors=context.config.throw_errors,
            debug=context.config.debug,
            original_exception=err,
        )

    return None



from typing import Dict, List

import numpy as np
from pyspark.sql import DataFrame, functions

from aporia.core.consts import TRAINING_BIN_COUNT
from aporia.core.errors import AporiaError
from aporia.core.types.field import FieldType
from aporia.training.api.log_training import FieldTrainingData

MEDIAN_QUANTILE = 0.5


def calculate_dataframe_training_data(
    data: DataFrame, fields_schema: Dict[str, FieldType]
) -> List[FieldTrainingData]:
    """Calculates training data for a PySpark dataframe.

    Args:
        data: PySpark DataFrame.
        fields_schema: Fields schema for the category that is being calculated.

    Returns:
        Training data for all of the fields in the dataframe.
    """
    training_data = []
    for field_name in data.columns:
        # Ignore fields that are not defined in the model version schema
        if field_name not in fields_schema:
            continue

        training_data.append(
            calculate_pyspark_training_data(
                field_name=field_name,
                field_data=data.select(field_name),
                field_type=fields_schema[field_name],
            )
        )

    return training_data


def calculate_pyspark_training_data(
    field_name: str, field_data: DataFrame, field_type: FieldType
) -> FieldTrainingData:
    """Calculates training data for a single field in a PySpark DataFrame.

    Args:
        field_name: Field name
        field_data: Field data
        field_type: Field type

    Returns:
        Field training data.
    """
    # We currently don't support datetime, vector & text training data
    if field_type in [FieldType.DATETIME, FieldType.VECTOR, FieldType.TEXT]:
        return _calculate_training_data_without_histogram(
            field_name=field_name, field_data=field_data
        )
    elif field_type == FieldType.NUMERIC:
        return _calculate_numeric_training_data(field_name=field_name, field_data=field_data)
    elif field_type in [FieldType.BOOLEAN, FieldType.STRING, FieldType.CATEGORICAL]:
        return _calculate_categorical_training_data(
            field_name=field_name, field_data=field_data, field_type=field_type
        )

    raise AporiaError("Unsupported field type {} of field {}".format(field_type.value, field_name))


def _calculate_training_data_without_histogram(
    field_name: str, field_data: DataFrame
) -> FieldTrainingData:
    # Discard None and infinite values
    valid_values = field_data.replace([np.inf, np.NINF], None).dropna()

    field_data_count = field_data.count()
    valid_values_count = valid_values.count()

    return FieldTrainingData(
        field_name=field_name,
        num_samples=valid_values_count,
        num_missing_values=field_data_count - valid_values_count,
    )


def _calculate_categorical_training_data(
    field_name: str, field_data: DataFrame, field_type: FieldType
) -> FieldTrainingData:
    # Discard None and infinite values
    valid_values = field_data.replace([np.inf, np.NINF], None).dropna()
    if field_type == field_type.BOOLEAN:
        valid_values = valid_values.select(valid_values[field_name].cast("boolean"))

    distinct_values = valid_values.groupBy(field_name).count().orderBy(field_name)
    bins = []
    counts = []
    for row in distinct_values.collect():
        bins.append(row[field_name])
        counts.append(row["count"])

    field_data_count = field_data.count()
    valid_values_count = valid_values.count()

    return FieldTrainingData(
        field_name=field_name,
        bins=bins,
        counts=counts,
        num_samples=valid_values_count,
        num_missing_values=field_data_count - valid_values_count,
        num_unique_values=len(bins),
    )


def _calculate_numeric_training_data(field_name: str, field_data: DataFrame) -> FieldTrainingData:
    # Discard None and infinite values
    valid_values = field_data.replace([np.inf, np.NINF], None).dropna()

    # DataFrame.rdd returns an RDD (Resilient Distributed Dataset) of Rows, use a map
    # function to get an RDD of values.
    # More about RDDs: https://spark.apache.org/docs/latest/rdd-programming-guide.html
    valid_values_rdd = valid_values.rdd.map(lambda x: x[0])
    bins, counts = valid_values_rdd.histogram(TRAINING_BIN_COUNT)  # type: ignore

    median = valid_values.approxQuantile(
        col=field_name, probabilities=[MEDIAN_QUANTILE], relativeError=0
    )[0]

    field_data_aggs = field_data.agg(
        # For each of the following aggregations, we use a sum aggregation in conjunction with
        # a when-otherwise cluase that will return 1 if the condition is True and 0 otherwise
        functions.sum(
            functions.when(field_data[field_name].isin([None, np.nan]), 1).otherwise(0)
        ).alias("missing_values"),
        functions.sum(functions.when(field_data[field_name].isin([np.inf]), 1).otherwise(0)).alias(
            "posinf_values"
        ),
        functions.sum(functions.when(field_data[field_name].isin([np.NINF]), 1).otherwise(0)).alias(
            "neginf_values"
        ),
    ).collect()[0]

    valid_values_aggs = valid_values.agg(
        functions.max(field_name).alias("max"),
        functions.min(field_name).alias("min"),
        functions.sum(field_name).alias("sum"),
        functions.avg(field_name).alias("avg"),
        functions.stddev(field_name).alias("std"),
        functions.variance(field_name).alias("var"),
        # For each of the following aggregations, we use a sum aggregation in conjunction with
        # a when-otherwise cluase that will return 1 if the condition is True and 0 otherwise
        functions.sum(functions.when(valid_values[field_name] == 0, 1).otherwise(0)).alias(
            "zero_values"
        ),
    ).collect()[0]

    return FieldTrainingData(
        field_name=field_name,
        bins=bins,
        counts=counts,
        min=valid_values_aggs["min"],
        max=valid_values_aggs["max"],
        sum=valid_values_aggs["sum"],
        median=median,
        average=valid_values_aggs["avg"],
        std=valid_values_aggs["std"],
        variance=valid_values_aggs["var"],
        num_samples=valid_values.count(),
        num_missing_values=field_data_aggs["missing_values"],
        num_posinf_values=field_data_aggs["posinf_values"],
        num_neginf_values=field_data_aggs["neginf_values"],
        num_zero_values=valid_values_aggs["zero_values"],
    )



from enum import Enum
from typing import Union

from aporia.core.context import get_context
from aporia.core.core_api import safe_api_function


class MonitorType(Enum):
    """The monitors implemented by Aporia."""

    MODEL_ACTIVITY = "model_activity"
    MISSING_VALUES = "missing_values"
    DATA_DRIFT = "data_drift"
    PREDICTION_DRIFT = "prediction_drift"
    VALUES_RANGE = "values_range"
    NEW_VALUES = "new_values"
    MODEL_STALENESS = "model_staleness"
    PERFORMANCE_DEGRADATION = "performance_degradation"
    METRIC_CHANGE = "metric_change"
    CUSTOM_METRIC = "custom_metric"


@safe_api_function("Creating monitor failed, error: {}")
def create_monitor(
    name: str,
    monitor_type: Union[MonitorType, str],
    scheduling: str,
    configuration: dict,
    is_active: bool = True,
    custom_alert_description: str = None,
) -> str:
    """Creates a new monitor.

    Args:
        name: A name for the new monitor, which will be displayed in Aporia's dashboard
        monitor_type: The type of monitor to create
        scheduling: A cron expression that indicates how often the monitor will run
        configuration: The monitor's configuration
        is_active: Whether or not the new monitor should be created as active
        custom_alert_description: A custom description for the monitor's alerts

    Returns:
        Created Monitor ID.
    """
    context = get_context()

    response = context.event_loop.run_coroutine(
        context.http_client.post(
            url="/monitors",
            data={
                "name": name,
                "type": MonitorType(monitor_type).value,
                "scheduling": scheduling,
                "configuration": configuration,
                "is_active": is_active,
                "custom_alert_description": custom_alert_description,
            },
        )
    )

    return response["id"]


@safe_api_function("Deleting monitor failed, error: {}")
def delete_monitor(monitor_id: str):
    """Deletes a monitor.

    Args:
        monitor_id: ID of the monitor to delete
    """
    context = get_context()
    context.event_loop.run_coroutine(context.http_client.delete(url=f"/monitors/{monitor_id}"))


@safe_api_function("Running monitors failed, error: {}")
def run_all_monitors(model_id: str, model_version: str):
    """Runs all monitors that are configured to run on the specified model version.

    Args:
        model_id: Model identifier, as received from the Aporia dashboard.
        model_version: Model version name.
    """
    query = """
        mutation RunAllMonitors(
            $modelId: String!,
            $modelVersion: String!,
            $environment: String!,
        ) {
            runAllMonitors(
                modelId: $modelId,
                modelVersion: $modelVersion,
                environment: $environment,
            ) {
                acknowledged
            }
        }
    """

    context = get_context()
    context.event_loop.run_coroutine(
        context.http_client.graphql(
            query,
            variables={
                "modelId": model_id,
                "modelVersion": model_version,
                "environment": context.config.environment,
            },
            graphql_url="/v1/tenant-hasura/v1/graphql",
        )
    )



from http import HTTPStatus
import logging
import re
import ssl
from typing import Optional
from urllib.parse import urljoin

import aiohttp
import certifi
import orjson
from tenacity import AsyncRetrying, retry_if_exception, stop_after_attempt, wait_exponential

from .errors import AporiaError, AporiaHTTPError
from .logging_utils import LOGGER_NAME
from .utils import orjson_serialize_default_handler

logger = logging.getLogger(LOGGER_NAME)


REQUEST_MAX_ATTEMPTS = 4
REQUEST_TIMEOUT = 10
REQUEST_RETRY_INITIAL_SLEEP = 2
DEFAULT_TIMEOUT_SEC = 30
GRAPHQL_API_PATH = "/v1/controller/graphql"
REST_API_BASE_PATH = "/v1beta"
HTTP_ERROR_MESSAGES_BY_STATUS = {
    HTTPStatus.UNAUTHORIZED: "Authentication failed, please check your token.",
    HTTPStatus.NOT_FOUND: "Aporia server not found, please check the 'host' and 'token' parameters you passed to aporia.init()",  # noqa: B950
    HTTPStatus.METHOD_NOT_ALLOWED: "Communication with Aporia server failed, please check your token.",
}
HTTP_502_VERBOSE_ERROR = """The server couldn't handle the rate at which you are sending requests.
Please try one of the following solutions:
    * Reduce the amount of log_* calls per second
    * Upgrade your cluster so it can support a higher load.
"""


class HttpClient:
    """Asynchronous http client."""

    def __init__(self, token: str, host: str, port: int, default_timeout: int):
        """Initialize a HttpClient instance.

        Args:
            token: Authorization token
            host: Aporia server address
            port: Aporia server port
            default_timeout: Default timeout
        """
        logger.debug("Initializing http client.")
        if not host.startswith("http"):
            host = f"https://{host}"

        self.base_url = f"{host}:{port}"
        self.ssl_context = ssl.create_default_context(cafile=certifi.where())

        self.headers = {"Authorization": f"Bearer {token}", "Content-Type": "application/json"}
        self.session = None
        self.default_timeout = default_timeout

    async def open(self):
        """Opens the HTTP session."""
        logger.debug("Creating HTTP session with Aporia server.")
        self.session = aiohttp.ClientSession(headers=self.headers)

    async def graphql(
        self,
        query: str,
        variables: dict,
        timeout: Optional[int] = None,
        retry_on_failure: bool = True,
        graphql_url: Optional[str] = None,
    ) -> dict:
        """Executes a GraphQL query and returns the result.

        Args:
            query: GraphQL query string
            variables: Variables for the query
            timeout: Timeout for the entire request, in seconds.
            retry_on_failure: True if the request should be retried on failure
                due to timeout or connection issues.
            graphql_url: GraphQL API URL.

        Returns:
            GraphQL query result
        """
        logger.debug("Sending GraphQL query", extra={"query": query, "variables": variables})

        data = {"query": query, "variables": variables}
        result = await self.request(
            method="POST",
            url=GRAPHQL_API_PATH if graphql_url is None else graphql_url,
            data=data,
            timeout=timeout,
            retry_on_failure=retry_on_failure,
        )

        # This can only happen if we send a graphql request to something that doesn't implement
        # a GraphQL API - this check is mostly here to make the linter happy
        if result is None:
            raise AporiaError("Unexpected error, GraphQL API response was empty.")

        # GraphQL always returns 200, making difficult to catch GraphQL errors in the
        # generic error handling logic - we handle it here explicitly instead
        if "errors" in result:
            self._handle_http_error(status=HTTPStatus.BAD_REQUEST, content=result)

        return result["data"]

    async def delete(
        self,
        url: str,
        timeout: Optional[int] = None,
        retry_on_failure: bool = True,
    ) -> Optional[dict]:
        """Performs a DELETE HTTP request (see HttpClient.request())."""
        return await self.request(
            method="DELETE",
            url=f"{REST_API_BASE_PATH}/{url}",
            timeout=timeout,
            retry_on_failure=retry_on_failure,
        )

    async def get(
        self,
        url: str,
        timeout: Optional[int] = None,
        retry_on_failure: bool = True,
    ) -> Optional[dict]:
        """Performs a GET HTTP request (see HttpClient.request())."""
        return await self.request(
            method="GET",
            url=f"{REST_API_BASE_PATH}/{url}",
            timeout=timeout,
            retry_on_failure=retry_on_failure,
        )

    async def post(
        self,
        url: str,
        data: dict,
        timeout: Optional[int] = None,
        retry_on_failure: bool = True,
    ) -> Optional[dict]:
        """Performs a POST HTTP request (see HttpClient.request())."""
        return await self.request(
            method="POST",
            url=f"{REST_API_BASE_PATH}/{url}",
            data=data,
            timeout=timeout,
            retry_on_failure=retry_on_failure,
        )

    async def request(  # type: ignore
        self,
        method: str,
        url: str,
        data: Optional[dict] = None,
        timeout: Optional[int] = None,
        retry_on_failure: bool = True,
    ) -> Optional[dict]:
        """Performs an HTTP request.

        Args:
            method: Request method string ("GET", "POST", etc.).
            url: Request URL.
            data: Optional request data, must be JSON serializable.
            timeout: Request timeout.
            retry_on_failure: True if the request should be retried on failure
                due to timeout or connection issues.

        Returns:
            The content of the HTTP response.
        """
        num_attempts = REQUEST_MAX_ATTEMPTS if retry_on_failure else 1
        timeout = timeout if timeout is not None else self.default_timeout
        async for attempt in AsyncRetrying(
            stop=stop_after_attempt(num_attempts),
            wait=wait_exponential(min=REQUEST_RETRY_INITIAL_SLEEP),
            retry=retry_if_exception(lambda err: not isinstance(err, AporiaError)),
            reraise=True,
        ):
            with attempt:
                return await self._request(method=method, url=url, data=data, timeout=timeout)

    async def _request(
        self, method: str, url: str, data: Optional[dict] = None, timeout: Optional[int] = None
    ) -> Optional[dict]:
        logger.debug(f"Sending {method} request to {url}.")

        # Strip redundant '/' from the URL
        normalized_url = re.sub("//+", "/", url)

        # Note: By default, aiohttp uses SSL and verifies the certificate in HTTPS requests
        async with self.session.request(  # type: ignore
            method=method,
            url=urljoin(self.base_url, normalized_url),
            data=None if data is None else self._serialize(data),
            timeout=timeout if timeout is not None else self.default_timeout,
            ssl=self.ssl_context,
        ) as response:
            response_content = await response.json()

            # All 2XX status codes (OK, CREATED, etc.) are considered success responses.
            # 3XX status codes are considered failures because we don't really handle redirects,
            # proxies, etc. All other statuses are obviously errors.
            if response.status < 200 or response.status >= 300:
                # This always raises an exception
                self._handle_http_error(
                    status=HTTPStatus(response.status), content=response_content
                )

            return response_content

    @staticmethod
    def _handle_http_error(status: HTTPStatus, content: Optional[dict] = None):
        short_message = None
        verbose_message = None

        if content is not None:
            # GraphQL error format
            if "errors" in content:
                short_message = content["errors"][0]["message"]
                verbose_message = content["errors"][0].get("extensions", {}).get("verbose_message")
            # REST API error format
            elif "error" in content:
                short_message = content["error"]["message"]
                verbose_message = content["error"].get("detail")

        # If error details were provided, raise an exception using those details
        # This should cover all errors that originate from our server.
        if short_message is not None:
            raise AporiaHTTPError(
                short_message=short_message,
                verbose_message=verbose_message,
                http_status=status,
            )

        # Otherwise, use a generic message for each exception type
        # This should cover authentication and routing errors.
        elif status in HTTP_ERROR_MESSAGES_BY_STATUS:
            raise AporiaHTTPError(
                short_message=HTTP_ERROR_MESSAGES_BY_STATUS[status],
                http_status=status,
            )

        # Bad gateway requires a more verbose error message, so we handle it separately
        elif status == HTTPStatus.BAD_GATEWAY:
            raise AporiaHTTPError(
                short_message="HTTP Gateway received invalid response from server",
                verbose_message=HTTP_502_VERBOSE_ERROR,
                http_status=status,
            )
        else:
            raise AporiaHTTPError(
                short_message=f"Unexpected HTTP error {status}",
                http_status=status,
            )

    @staticmethod
    def _serialize(data: dict) -> bytes:
        try:
            return orjson.dumps(
                data,
                option=orjson.OPT_SERIALIZE_NUMPY,
                default=orjson_serialize_default_handler,
            )
        except TypeError as err:
            raise AporiaError(str(err))

    async def close(self):
        """Closes the http session."""
        await self.session.close()



from typing import Dict, List

import numpy as np
from pandas import DataFrame, Series

from aporia.core.consts import TRAINING_BIN_COUNT
from aporia.core.errors import AporiaError
from aporia.core.types.field import FieldType
from .api.log_training import FieldTrainingData


def calculate_dataframe_training_data(
    data: DataFrame, fields_schema: Dict[str, FieldType]
) -> List[FieldTrainingData]:
    """Calculates training data for all fields in a DataFrame.

    Args:
        data: Raw training data
        fields_schema: Fields schema for the category that is being calculated.

    Returns:
        Training data for all fields in the dataframe
    """
    training_data = []
    for field_name, field_data in data.items():
        # Ignore fields that are not defined in the model version schema
        if field_name not in fields_schema:
            continue

        training_data.append(
            calculate_training_data(
                field_name=field_name,
                field_data=field_data,
                field_type=fields_schema[field_name],
            )
        )

    return training_data


def calculate_training_data(
    field_name: str, field_data: Series, field_type: FieldType
) -> FieldTrainingData:
    """Calculates training data for a single field.

    Args:
        field_name: Field name
        field_data: Field data
        field_type: Field type

    Returns:
        Field training data.
    """
    # We currently don't support datetime, vector & text training data
    if field_type in [FieldType.DATETIME, FieldType.VECTOR, FieldType.TEXT]:
        return _calculate_training_data_without_histogram(
            field_name=field_name, field_data=field_data
        )
    elif field_type == FieldType.NUMERIC:
        return _calculate_numeric_training_data(field_name=field_name, field_data=field_data)
    elif field_type in [FieldType.BOOLEAN, FieldType.STRING, FieldType.CATEGORICAL]:
        return _calculate_categorical_training_data(
            field_name=field_name, field_data=field_data, field_type=field_type
        )

    raise AporiaError("Unsupported field type {} of field {}".format(field_type.value, field_name))


def _calculate_training_data_without_histogram(
    field_name: str, field_data: Series
) -> FieldTrainingData:
    valid_values = field_data[field_data.notnull()]

    return FieldTrainingData(
        field_name=field_name,
        num_samples=len(valid_values),
        num_missing_values=len(field_data) - len(valid_values),
    )


def _calculate_categorical_training_data(
    field_name: str, field_data: Series, field_type: FieldType
) -> FieldTrainingData:
    valid_values = field_data[field_data.notnull()]
    if field_type == field_type.BOOLEAN:
        valid_values = valid_values.astype("bool")

    bins, counts = np.unique(valid_values, return_counts=True)

    # Note: There is a possible edge case here in which a user passes an infinite value
    # as one of the categories. We chose not to count those values at the moment, since
    # most numpy functions don't handle str and bool well, which would force us to split
    # this function up for each field type.
    return FieldTrainingData(
        field_name=field_name,
        bins=bins.tolist(),
        counts=counts.tolist(),
        num_samples=len(valid_values),
        num_missing_values=len(field_data) - len(valid_values),
        num_unique_values=len(bins),
    )


def _calculate_numeric_training_data(field_name: str, field_data: Series) -> FieldTrainingData:
    # Cast everything to float and normalize null values (some operations don't work on pd.NA)
    field_data = field_data.fillna(np.nan).astype(float)  # type: ignore

    # Filter out infinite and NaN values (isfinite is False for NaN)
    finite_values = field_data[np.isfinite(field_data)]

    counts, bins = np.histogram(finite_values, bins=TRAINING_BIN_COUNT)

    return FieldTrainingData(
        field_name=field_name,
        bins=bins.tolist(),
        counts=counts.tolist(),
        min=np.min(finite_values),
        max=np.max(finite_values),
        sum=np.sum(finite_values),  # type: ignore
        median=np.median(finite_values),
        average=np.average(finite_values),
        std=np.std(finite_values),  # type: ignore
        variance=np.var(finite_values),  # type: ignore
        num_samples=len(finite_values),
        num_missing_values=np.count_nonzero(np.isnan(field_data)),
        num_posinf_values=np.count_nonzero(np.isposinf(field_data)),
        num_neginf_values=np.count_nonzero(np.isneginf(field_data)),
        num_zero_values=np.count_nonzero(finite_values == 0),
    )



from functools import lru_cache
import logging
from typing import cast, Dict, List, Optional, Tuple, Union

from numpy import nan, ndarray
from pandas import DataFrame

from aporia.core.base_model import BaseModel, validate_model_ready
from aporia.core.logging_utils import LOGGER_NAME
from aporia.core.types.field import FieldCategory, FieldType
from aporia.core.utils import generate_features_schema_from_shape
from .api.get_model_version import get_model_version
from .api.log_training import (
    FieldTrainingData,
    log_test_data,
    log_training_data,
    log_training_sample_data,
)
from .training import calculate_dataframe_training_data
from .training_validator import TrainingDataset, TrainingValidator

logger = logging.getLogger(LOGGER_NAME)


SAMPLE_SIZE = 1000


class TrainingModel(BaseModel):
    """Model object for logging training events."""

    def __init__(self, model_id: str, model_version: str):
        """Initializes a TrainingModel object.

        Args:
            model_id: Model identifier, as received from the Aporia dashboard.
            model_version: Model version - this can be any string that represents the model
                version, such as "v1" or a git commit hash.
        """
        super().__init__(model_id=model_id, model_version=model_version, set_ready_when_done=True)

    @validate_model_ready
    def log_training_set(
        self,
        features: Union[DataFrame, ndarray],
        labels: DataFrame,
        raw_inputs: Optional[DataFrame] = None,
    ):
        """See aporia.model.Model."""
        logger.debug("Logging training set.")
        with self.handle_error("Logging training set failed, error: {}"):
            version_schema = self.get_version_schema()
            validator = TrainingValidator(
                dataset_type=TrainingDataset.TRAINING, schema=version_schema
            )
            validator.validate(features=features, labels=labels, raw_inputs=raw_inputs)

            raw_inputs_training_data = None
            if raw_inputs is not None:
                raw_inputs_training_data = calculate_dataframe_training_data(
                    raw_inputs, version_schema[FieldCategory.RAW_INPUTS]
                )

            if isinstance(features, DataFrame):
                features_dataframe = features
                features_schema = version_schema[FieldCategory.FEATURES]
            else:
                features_dataframe, features_schema = self._build_features_dataframe_and_schema(
                    features=features
                )

            self.log_training_sample(features=features_dataframe, labels=labels)
            self.log_training_set_aggregations(
                features=calculate_dataframe_training_data(features_dataframe, features_schema),
                labels=calculate_dataframe_training_data(
                    labels, version_schema[FieldCategory.PREDICTIONS]
                ),
                raw_inputs=raw_inputs_training_data,
            )

    @validate_model_ready
    def log_test_set(
        self,
        features: Union[DataFrame, ndarray],
        predictions: DataFrame,
        labels: DataFrame,
        raw_inputs: Optional[DataFrame] = None,
        confidences: Optional[ndarray] = None,
    ):
        """See aporia.model.Model."""
        logger.debug("Logging test set.")
        with self.handle_error("Logging test set failed, error: {}"):
            version_schema = self.get_version_schema()
            validator = TrainingValidator(dataset_type=TrainingDataset.TEST, schema=version_schema)
            validator.validate(
                features=features, labels=labels, predictions=predictions, raw_inputs=raw_inputs
            )

            raw_inputs_training_data = None
            if raw_inputs is not None:
                raw_inputs_training_data = calculate_dataframe_training_data(
                    raw_inputs, version_schema[FieldCategory.RAW_INPUTS]
                )

            if isinstance(features, DataFrame):
                features_dataframe = features
                features_schema = version_schema[FieldCategory.FEATURES]
            else:
                features_dataframe, features_schema = self._build_features_dataframe_and_schema(
                    features=features
                )

            self.log_test_set_aggregations(
                features=calculate_dataframe_training_data(features_dataframe, features_schema),
                predictions=calculate_dataframe_training_data(
                    predictions, version_schema[FieldCategory.PREDICTIONS]
                ),
                labels=calculate_dataframe_training_data(
                    labels, version_schema[FieldCategory.PREDICTIONS]
                ),
                raw_inputs=raw_inputs_training_data,
            )

    @validate_model_ready
    def log_training_set_aggregations(
        self,
        features: List[FieldTrainingData],
        labels: List[FieldTrainingData],
        raw_inputs: Optional[List[FieldTrainingData]] = None,
    ):
        """Logs training set training aggregations.

        Args:
            features: Feature field aggregations
            labels: Label (prediction) field aggregations
            raw_inputs: Raw Inputs aggregations
        """
        self._event_loop.run_coroutine(
            log_training_data(
                http_client=self._http_client,
                model_id=self.model_id,
                model_version=self.model_version,
                features=features,
                labels=labels,
                raw_inputs=raw_inputs,
            )
        )

    @validate_model_ready
    def log_test_set_aggregations(
        self,
        features: List[FieldTrainingData],
        predictions: List[FieldTrainingData],
        labels: List[FieldTrainingData],
        raw_inputs: Optional[List[FieldTrainingData]] = None,
    ):
        """Logs test set training aggregations.

        Args:
            features: Feature field aggregations
            predictions: Prediction field aggregations
            labels: Label (prediction) field aggregations
            raw_inputs: Raw Inputs aggregations
        """
        self._event_loop.run_coroutine(
            log_test_data(
                http_client=self._http_client,
                model_id=self.model_id,
                model_version=self.model_version,
                features=features,
                predictions=predictions,
                labels=labels,
                raw_inputs=raw_inputs,
            )
        )

    @validate_model_ready
    def log_training_sample(
        self,
        features: Union[DataFrame, ndarray],
        labels: DataFrame,
    ):
        """Logs training sample.

        Args:
            features: Training set features
            labels: Training set labels

        Notes:
            * As this is a sample for analytical purpose, we log a maximum of 1000 rows.
        """
        if isinstance(features, ndarray):
            features, _ = self._build_features_dataframe_and_schema(features)

        # Null values are replaced with np.nan, because orjson can't serialize pd.NA
        features_sample = cast(
            DataFrame, features.sample(min(SAMPLE_SIZE, len(features))).fillna(nan)
        )
        labels_sample = cast(DataFrame, labels.sample(min(SAMPLE_SIZE, len(labels))).fillna(nan))

        self._event_loop.run_coroutine(
            log_training_sample_data(
                http_client=self._http_client,
                model_id=self.model_id,
                model_version=self.model_version,
                features=features_sample.to_dict("records"),
                labels=labels_sample.to_dict("records"),
            )
        )

    @lru_cache(maxsize=1)
    @validate_model_ready
    def get_version_schema(self) -> Dict[FieldCategory, Dict[str, FieldType]]:
        """Fetches the schema of this model version.

        Returns:
            Model version schema
        """
        return self._event_loop.run_coroutine(
            get_model_version(
                http_client=self._http_client,
                model_id=self.model_id,
                model_version=self.model_version,
            )
        )

    @staticmethod
    def _build_features_dataframe_and_schema(
        features: ndarray,
    ) -> Tuple[DataFrame, Dict[str, FieldType]]:
        features_schema = generate_features_schema_from_shape(features.shape[1:])
        for feature_name in features_schema:
            features_schema[feature_name] = FieldType(features_schema[feature_name])

        flattened_data = features.reshape(features.shape[0], len(features_schema))
        features_dataframe = DataFrame(
            data=flattened_data, columns=list(features_schema.keys()), copy=False
        )

        return features_dataframe, features_schema



from collections import abc, OrderedDict
from typing import Dict, Optional, Union

import numpy as np
import pandas as pd

from aporia.core.context import get_context
from aporia.core.errors import AporiaError, handle_error
from aporia.core.types.field import FieldType, FieldValue

# This is based on https://numpy.org/doc/stable/reference/generated/numpy.dtype.kind.html#numpy.dtype.kind
DTYPE_TO_FIELD_TYPE = {
    "b": FieldType.BOOLEAN,
    "i": FieldType.NUMERIC,
    "u": FieldType.NUMERIC,
    "f": FieldType.NUMERIC,
    "M": FieldType.DATETIME,
    "O": FieldType.STRING,
    "S": FieldType.STRING,
    "U": FieldType.STRING,
}

NUMERIC_DTYPES = ["i", "u", "f"]
STRING_DTYPES = ["O", "S", "U"]
STRING_UNIQUE_RATIO = 0.25
MIN_UNIQUE_VALUES_FOR_TEXT = 50


def infer_type_from_dtype_and_data(
    dtype: np.dtype, data: Optional[pd.Series]
) -> Optional[FieldType]:
    """Attempts to convert a numpy/pandas dtype to a FieldType.

    Args:
        dtype: Dtype to convert
        data: Numpy data series

    Returns:
        FieldType that matches the dtype, or None if conversion is impossible
    """
    if isinstance(dtype, pd.api.types.CategoricalDtype):
        category_type = DTYPE_TO_FIELD_TYPE.get(dtype.categories.dtype.kind)
        # We only support categorical fields with numeric categories
        if category_type == FieldType.NUMERIC:
            return FieldType.CATEGORICAL

        return category_type

    if data is not None and not data.empty:
        if dtype.kind in NUMERIC_DTYPES and isinstance(data.values[0], abc.Iterable):
            return FieldType.VECTOR

        # We need to decided if a string type is str or text
        # We check if there are more than 50 unique values
        # which are more than 25% of the total values
        if dtype.kind in STRING_DTYPES and len(data.shape) > 0:
            _, unique_counts = np.unique(data.values, return_counts=True)
            uniques = len(unique_counts)
            if (
                uniques > MIN_UNIQUE_VALUES_FOR_TEXT
                and uniques > data.shape[0] * STRING_UNIQUE_RATIO
            ):
                return FieldType.TEXT

    return DTYPE_TO_FIELD_TYPE.get(dtype.kind)  # type: ignore


def pandas_to_dict(data: Union[pd.DataFrame, pd.Series]) -> Optional[Dict[str, FieldValue]]:
    """Converts a pandas DataFrame or Series to a dict for log_* functions.

    Args:
        data: DataFrame or Series to convert.

    Returns:
        The data converted to a dict, mapping field names to their values

    Notes:
        * data must contain column names that match the fields defined in create_model_version
        * If data is a DataFrame, it must contain exactly one row
    """
    context = None
    try:
        context = get_context()

        if isinstance(data, pd.Series):
            return data.fillna(np.nan).to_dict()  # type: ignore
        elif isinstance(data, pd.DataFrame):
            num_rows, _ = data.shape
            if num_rows > 1:
                raise AporiaError("cannot convert DataFrame with more than 1 row")

            return data.iloc[0].fillna(np.nan).to_dict()
        else:
            raise AporiaError("data must be a pandas DataFrame or Series")

    except Exception as err:
        handle_error(
            message_format="Converting pandas data to dict failed, {}",
            verbose=False if context is None else context.config.verbose,
            throw_errors=False if context is None else context.config.throw_errors,
            debug=False if context is None else context.config.debug,
            original_exception=err,
        )

    return None


def infer_schema_from_dataframe(data: pd.DataFrame) -> Optional[Dict[str, str]]:
    """Infers model version schema from a pandas DataFrame or Series.

    Field names and types are inferred from column names and types.

    Args:
        data: pandas DataFrame or Series

    Returns:
        A schema describing the data, as required by the create_model_version function.

    Notes:
        * The field types are inferred using the following logic, based on the column dtypes:
            * dtype="category" with numeric (integer of float) categories -> categorical field
            * dtype="category" with non-numeric categories -> See rules below
            * Array of numeric values (integer or float) -> vector field
            * dtype="bool" -> boolean field
            * dtypes that represent signed/unsigned integers and floating point numbers -> numeric field
            * dtype is "string", "unicode", "object" -> string field
            * dtype is "string", "unicode", "object", more than 50 values which more than
                25% of them are unique -> text field
            * dtype is any datetime type (with or without timezone) -> datetime field
        * If data contains a column with a type that doesn't match any of the rules
          described above, an error will be raised.
    """
    context = get_context()
    try:
        if not isinstance(data, pd.DataFrame):
            raise AporiaError(
                "cannot infer schema from {}, data must be a pandas DataFrame".format(type(data))
            )

        schema = OrderedDict()
        for column_name, values in data.items():
            values_without_nulls = values.dropna().infer_objects()
            column_type = infer_type_from_dtype_and_data(
                values_without_nulls.dtype, values_without_nulls
            )
            if column_type is None:
                raise AporiaError(
                    "the dtype {} of column {} is not supported".format(
                        values_without_nulls.dtype, column_name
                    )
                )

            schema[column_name] = column_type.value

        return schema

    except Exception as err:
        handle_error(
            message_format="Inferring schema from dataframe failed, {}",
            verbose=context.config.verbose,
            throw_errors=context.config.throw_errors,
            debug=context.config.debug,
            original_exception=err,
        )

    return None
